"""Tests for OpenStack API twisted client"""

from twisted.internet import (
    defer,
    reactor,
    )
from twisted.python.failure import Failure
from twisted.web import (
    http_headers,
    client as txclient,
    )

from juju import errors
from juju.lib import (
    mocker,
    testing,
    )

from juju.providers.openstack import (
    client,
    credentials,
    )


class StubClient(client._OpenStackClient):

    def __init__(self):
        self.url = "http://testing.invalid"

    make_url = client._OpenStackClient._make_url


class TestMakeUrl(testing.TestCase):

    def setUp(self):
        self.client = StubClient()
        self.client.services = {
            "compute": self.client.url + "/nova",
            "object-store": self.client.url + "/swift",
            }

    def test_list_str(self):
        self.assertEqual("http://testing.invalid/nova/servers",
            self.client.make_url("compute", ["servers"]))
        self.assertEqual("http://testing.invalid/swift/container/object",
            self.client.make_url("object-store", ["container", "object"]))

    def test_list_int(self):
        self.assertEqual("http://testing.invalid/nova/servers/1000",
            self.client.make_url("compute", ["servers", 1000]))
        self.assertEqual("http://testing.invalid/nova/servers/1000/detail",
            self.client.make_url("compute", ["servers", 1000, "detail"]))

    def test_list_unicode(self):
        url = self.client.make_url("object-store", ["container", u"\xa7"])
        self.assertIsInstance(url, str)
        self.assertEqual("http://testing.invalid/swift/container/%C2%A7", url)

    def test_str(self):
        self.assertEqual("http://testing.invalid/nova/servers",
            self.client.make_url("compute", "servers"))
        self.assertEqual("http://testing.invalid/swift/container/object",
            self.client.make_url("object-store", "container/object"))

    def test_trailing_slash(self):
        self.client.services["object-store"] += "/"
        self.assertEqual("http://testing.invalid/nova/container",
            self.client.make_url("compute", "container"))
        self.assertEqual("http://testing.invalid/nova/container/object",
            self.client.make_url("compute", ["container", "object"]))


class FakeResponse(object):
    """Bare minimum needed to look like a twisted http response"""

    def __init__(self, code, headers):
        self.code = code
        self.headers = headers
        self.length = 0


class ClientTests(testing.TestCase):
    """Testing of low level client behaviour

    Rough temporary tests until client rearrangements make this easier.
    """

    class Checker(object):
        """Standin for cert checker that exists regardless of txaws"""

    class SSLError(Exception):
        """Standin for ssl exception that exists regardless of OpenSSL"""

    def setUp(self):
        super(ClientTests, self).setUp()
        self.patch(client, "SSLError", self.SSLError)
        self.patch(client, "WebVerifyingContextFactory", self.Checker)
        self.mock_agent = self.mocker.replace("twisted.web.client.Agent",
            passthrough=False)
        self.mocker.order()

    def get_credentials(self, config):
        return credentials.OpenStackCredentials(config)

    def is_checker(self, obj):
        return isinstance(obj, self.Checker)

    def make_client_legacy(self):
        config = {
            "auth-url": "https://testing.invalid",
            "auth-mode": "legacy",
            "username": "user",
            "access-key": "key",
            }
        osc = client._OpenStackClient(self.get_credentials(config), True)
        self.mock_agent(reactor, contextFactory=mocker.MATCH(self.is_checker))
        self.mocker.result(self.mock_agent)
        return config, osc

    @defer.inlineCallbacks
    def test_auth_legacy(self):
        config, osc = self.make_client_legacy()
        # TODO: check headers for correct auth values
        self.mock_agent.request("GET", config["auth-url"], mocker.ANY, None)
        response = FakeResponse(204, http_headers.Headers({
            "X-Server-Management-Url": ["http://testing.invalid/compute"],
            "X-Auth-Token": ["tok"],
            }))
        self.mocker.result(defer.succeed(response))
        self.mocker.replay()
        log = self.capture_logging()
        yield osc.authenticate()
        self.assertEqual("tok", osc.token)
        self.assertEqual("http://testing.invalid/compute/path",
            osc._make_url("compute", ["path"]))
        self.assertIn("compute service not using secure", log.getvalue())

    def test_cert_failure(self):
        config, osc = self.make_client_legacy()
        self.mock_agent.request("GET", config["auth-url"], mocker.ANY, None)
        response = FakeResponse(204, http_headers.Headers({
            "X-Server-Management-Url": ["http://testing.invalid/compute"],
            "X-Auth-Token": ["tok"],
            }))
        self.mocker.result(defer.fail(txclient.ResponseFailed([
            Failure(self.SSLError()),
            ])))
        self.mocker.replay()
        deferred = osc.authenticate()
        return self.assertFailure(deferred, errors.SSLVerificationError)


class TestPlan(testing.TestCase):
    """Ideas for tests needed"""

    # auth request without auth
    # get bytes, content-length 0, return ""
    # get bytes, not ResponseDone, raise wrapped in ProviderError (with any bytes?)
    # get bytes, type json, return bytes
    # get json, content length 0, raise ProviderError
    # get json, bad header (several forms), raise ProviderError (with bytes)
    # get json, not ResponseDone, raise wrapped in ProviderError
    #    (with any bytes?)
    # get json, undecodable, raise wrapped in ProviderError with bytes
    # get json, mismatching root, raise ProviderError with bytes or json?
    # wrong code, no json header, raise ProviderError with bytes
    # wrong code, not ResponseDone, raise ProviderError from code
    #    with any bytes
    # wrong code, undecodable, raise ProviderError from code with bytes
    # wrong code, has mystery root, raise ProviderError from code with bytes or json?
    # wrong code, has good root, no message
    # wrong code, has good root, no code
    # wrong code, has good root, differing code
    # wrong code, has good root, message, and matching code
